/***************************************************************************
 *   Copyright (C) 2005 by Nicola Di Mauro                                 *
 *   ndm@di.uniba.it                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef _LISTAP_H
#define _LISTAP_H

#include "cellalp.h"
#include <iostream>
using namespace std;

template<class T>
class circLista
  {
  public:
    circLista();
    ~circLista();
    circLista(const circLista<T>&);
    void copia(const circLista<T>&);     
    /* posizione  un puntatore a cella */
    typedef Cella<T>* posizione;
    typedef T tipoelem;
    /* Prototipi degli operatori */
    void crealista();
    void cancella();
    bool listavuota() const;
    tipoelem leggilista(posizione) const;
    void scrivilista(tipoelem, posizione);
    posizione primoLista() const;
    bool finelista(posizione) const;
    posizione succlista(posizione) const;
    posizione preclista(posizione) const;
    void inslista(tipoelem,posizione&);
    void canclista(posizione&);
    void scambia(posizione&, posizione&);
    // funzioni di servizio
    void stampaLista();
    circLista<T> ordina();
    circLista<T> partiziona(tipoelem elem);
  private:
    posizione lista; //la lista  un puntatore ad oggetto Cella
  };

/* Liste: Rappresentazione collegata circolare (con sentinella)
 * realizzata mediante doppi puntatori (o simmetrica)
 */

template <class T> circLista<T>::circLista()
{crealista();}

template <class T> circLista<T>::circLista(const circLista<T> &sorgente )
{
   // cancella();
   crealista();
   Cella<T>* indice1=primoLista();
   Cella<T>* indice2=sorgente.primoLista();
   while (!sorgente.finelista(indice2))
   {
        T elem = sorgente.leggilista(indice2);
        // cout << "letto " << elem << "\n";
        indice1=succlista(indice1);
        inslista(elem,indice1);
        // cout << "stampalista dopo inserimento\n";
        indice2 = sorgente.succlista(indice2);          
   }
}

template <class T> void circLista<T>::copia(const circLista<T> &sorgente )
{
   cancella();
   crealista();
   Cella<T>* indice1=primoLista();
   Cella<T>* indice2=sorgente.primoLista();
   while (!sorgente.finelista(indice2))
   {
        T elem = sorgente.leggilista(indice2);
        // cout << "letto " << elem << "\n";
        indice1=succlista(indice1);
        inslista(elem,indice1);
        // cout << "stampalista dopo inserimento\n";
        indice2 = sorgente.succlista(indice2);          
   }
}

template <class T> void circLista<T>::cancella()
{
    while (lista->getSucc() != lista->getPrec()) 
    {
       Cella<T>* posizione = lista->getSucc();
       canclista(posizione);
    }   
    delete lista;
}

template <class T> circLista<T>::~circLista()
{
    while (lista->getSucc() != lista->getPrec()) 
    {
       Cella<T>* posizione = lista->getSucc();
       canclista(posizione);
    }   
    delete lista;
}

template <class T> void circLista<T>::crealista()
{
  T ElementoNullo;
  lista = new Cella<T>;
  lista->setElemento(ElementoNullo);
  lista->setSucc(lista);
  lista->setPrec(lista);
  //la sentinella punta a se stessa
}


template <class T> bool circLista<T>::listavuota() const
  {
    return ((lista->getSucc() == lista) &&  (lista->getPrec()==lista));
  }

template <class T> Cella<T>* circLista<T>::primoLista() const
  {
    return lista->getSucc();
  }

template <class T> Cella<T>* circLista<T>::succlista(posizione p) const
  {
    return p->getSucc();
  }

template <class T> Cella<T>* circLista<T>::preclista(posizione p) const
  {
    return p->getPrec();
  }

template <class T>  bool circLista<T>::finelista(posizione p) const
  {
    return (p==lista);
  }

template <class T> T circLista<T>::leggilista(posizione p) const
  {
    return p->getElemento();
  }

template <class T> void circLista<T>::scrivilista(tipoelem a, posizione p)
{
  p->setElemento(a);
}

template <class T> void circLista<T>::inslista(tipoelem a, posizione &p)
{  
  Cella<T>* temp = new Cella<T>;
  temp->setElemento(a);
  temp->setPrec(p->getPrec());
  temp->setSucc(p);
  (p->getPrec())->setSucc(temp);
  p->setPrec(temp);
  p=temp; // se p era la posizione dell'elemento n-mo, adesso lo  temp 
}

template <class T> void circLista<T>::canclista(posizione &p)
{
  Cella<T>* temp = new Cella<T>;
  temp=p;
  (p->getSucc())->setPrec(p->getPrec());
  (p->getPrec())->setSucc(p->getSucc());
  p=p->getSucc();
  delete(temp);
}

template <class T> void circLista<T>::stampaLista()
{
  cout<<"[";
  Cella<T>* indice=primoLista();
  while (!finelista(indice))
  {
      cout << leggilista(indice);
      if (!finelista(succlista(indice)))
        cout << ", ";
      indice = succlista(indice);
      
    }
  cout<<"]\n";
}

template <class T> circLista<T> circLista<T>::ordina()
{
  // ordinamento per inserzione con creazione nuova lista
  Cella<T>* indicenuova;
  Cella<T>* indice;
  circLista<T> nuova;
          
  if (!listavuota())
  {
     indicenuova=nuova.primoLista(); // indicenuova scandisce la lista ordinata
     indice=primoLista(); // indice scandisce la lista non ordinata
     indicenuova=succlista(nuova.primoLista());
     nuova.inslista(leggilista(indice),indicenuova); // ho trasferito il primo elemento in nuova
  }
  indice=succlista(indice); // parto dal secondo elemento nella lista non ordinata

    
  while (!finelista(indice)) // indice scandisce la lista non ordinata
  {
      T elem=leggilista(indice);
      indicenuova=nuova.primoLista();
      // cerca la posizione di inserimento      
      while ( !nuova.finelista(indicenuova) && elem>leggilista(indicenuova))
      {
            indicenuova=nuova.succlista(indicenuova);
      }
      nuova.inslista(elem,indicenuova); 
      indice=succlista(indice);
  } 
  return nuova; 
}

template <class T> void circLista<T>::scambia(posizione &a, posizione &b)
{  
  T buffer = leggilista(a);   
  scrivilista(leggilista(b),a);  
  scrivilista(buffer,b);
}

template <class T> circLista<T> circLista<T>::partiziona(tipoelem elem)
{  
  // da scrivere
}

#endif // _LISTAP_H 
